           (*-------------------------------------------*
            |                   Test                    |
            |        CSP-Prover on Isabelle2005         |
            |                  April 2006               |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Test_Buffer = CSP_F:

(*****************************************************************

         1. test Buffer
         2. verification of deadlock-free.
         3.
         4. 

 *****************************************************************)

datatype Event = In real | Out "real * nat"
datatype BufferName = Empty nat | Full real nat

consts 
  funDef :: "(BufferName, Event) procDef"

recdef funDef "{}"
  "funDef  (f, (Empty n))  = In ? r -> (f (Full r n))"
  "funDef  (f, (Full r n)) = Out (r,n) -> (f (Empty (Suc n)))"

syntax
  "_fun"    :: "(BufferName, Event) procFun"  ("fun")

translations
  "fun" == "curry funDef"

consts
  Buffer :: "Event proc"

defs
  Buffer_def: "Buffer == (FIX fun)(Empty 0)"

(*** Spc ***)

consts 
  DF :: "Event proc"

defs
  DF_def : "DF == (MU X .. (! x -> X))"

(*** To automatically unfold syntactic sugar ***)
declare csp_prefix_ss_def[simp]
declare inj_on_def[simp]

lemma ProcFun_fun[simp]: "fun : ProcFun"
apply (auto simp add: ProcFun_def)
apply (induct_tac p)
apply (auto)
done

(*********************************************************
            relation between Buffer and DF
 *********************************************************)

consts 
  Buffer_to_DF :: "BufferName => Event proc"

primrec
  "Buffer_to_DF (Empty n)  = DF"
  "Buffer_to_DF (Full r n) = DF"

(*********************************************************
               Buffer is deadlock free.
 *********************************************************)

(*** manual proof ***)

lemma manual_proof_Buffer: "DF <=F Buffer"
apply (simp add: Buffer_def)
apply (rule cspF_fp_induct_right[of _ _ "Buffer_to_DF"])
apply (simp_all add: DF_def)

apply (induct_tac p)
apply (simp_all add: DF_def)

(* Empty *)
apply (rule cspF_rw_left)
apply (rule cspF_unwind_MU)
apply (simp)

apply (rule cspF_decompo_subset)
apply (simp)
apply (simp)
apply (rule cspF_reflex)

(* Full *)
apply (rule cspF_rw_left)
apply (rule cspF_unwind_MU)
apply (simp)

apply (rule cspF_rw_right)
apply (rule cspF_step)

apply (rule cspF_decompo_subset)
apply (simp)
apply (simp)
apply (rule cspF_reflex)
done

(*** semi-automatic proof ***)

lemma semi_auto_proof_Buffer: "DF <=F Buffer"
apply (simp add: Buffer_def)
apply (rule cspF_fp_induct_right[of _ _ "Buffer_to_DF"])
apply (simp_all add: DF_def)

apply (induct_tac p)
apply (simp_all add: DF_def)
apply (tactic {* cspF_hsf_tac 1 *}, rule cspF_decompo_subset, simp_all)+
done

end
