           (*-------------------------------------------*
            |                   Test                    |
            |                                           |
            |        CSP-Prover on Isabelle2004         |
            |               August 2004                 |
            |             December 2004 (modified)      |
            |                                           |
            |        CSP-Prover on Isabelle2005         |
            |               November 2005  (modified)   |
            |                  April 2006  (modified)   |
            |                                           |
            |        Yoshinao Isobe (AIST JAPAN)        |
            *-------------------------------------------*)

theory Test_infinite = CSP_F:

(*****************************************************************

         1. simple example for fixed point inductuction theorem
         2. Parallel, Hiding, Internal choice
         3. Refinement
         4. 

 *****************************************************************)

(*********************************************************
                         event
 *********************************************************)

datatype Event = Num nat | Read nat
datatype SpcName = SPC nat
datatype ImpName = UI | VAR nat

(*********************************************************
            specification SPC and system IMP
 *********************************************************)

consts
  GTs :: "nat => nat set"

defs
  GTs_def : "GTs n == {m. n < m}"

(*** Spc ***)

consts 
  SpcDef :: "(SpcName, Event) procDef"

recdef SpcDef "{}"
  "SpcDef  (X, (SPC n)) = Num n -> (!nat m:(GTs n) .. (X (SPC m)))"

syntax
  "_SpcFun" :: "(SpcName, Event) procFun" ("SpcFun")
  "_SpcFix" :: "SpcName => Event proc"    ("SpcFix")

translations
    "SpcFun" == "curry SpcDef"
  "SpcFix n" == "(FIX! SpcFun) n"

consts 
  Spc :: "Event proc"

defs
  Spc_def: "Spc == SpcFix (SPC 0)"

(*** Imp ***)

consts 
  ImpDef :: "(ImpName, Event) procDef"

recdef ImpDef "{}"
  "ImpDef      (X, UI) = Read ? m -> Num m -> (X UI)"
  "ImpDef   (X, VAR n) = Read ! n -> (X (VAR (Suc n)))"

syntax
  "_ImpFun" :: "(ImpName, Event) procFun" ("ImpFun")
  "_ImpFix" :: "ImpName => Event proc"    ("ImpFix")

translations
    "ImpFun" == "curry ImpDef"
  "ImpFix n" == "(FIX! ImpFun) n"

consts 
  Imp :: "Event proc"

defs
  Imp_def: "Imp == (ImpFix UI |[range Read]| ImpFix (VAR 0)) -- (range Read)"

(*********************************************************
            relation between SPC and IMP
 *********************************************************)

consts 
  Spc_to_Imp :: "SpcName => Event proc"

primrec
  "Spc_to_Imp (SPC n) = (ImpFix UI |[range Read]| ImpFix (VAR n)) -- (range Read)"

(*********************************************************
                     small lemmas
 *********************************************************)

declare Int_commute [simp]

lemma set1[simp]: "range Read Int {Read n} = {Read n}"
by (auto)

lemma set2[simp]: "{Read (Suc n)} Int (range Read Int {Num n}) Un
                    ({Num n} - range Read) = {Num n}"
by (auto)

lemma set3[simp]: "Num n ~: range Read"
by (simp add: image_def)

(*********************************************************
          gProc lemmas (rutine work)
 *********************************************************)

(*** To automatically unfold syntactic sugar ***)
declare csp_prefix_ss_def[simp]

lemma gProcFun_Spc_Imp[simp]:
      "SpcFun : gProcFun"
      "ImpFun : gProcFun"
by (simp add: CSP_Fun_def, rule allI, induct_tac p, simp_all)+

declare inj_on_def[simp]

(*********************************************************
                   ? SPC <=F IMP ?
 *********************************************************)

lemma "Spc <=F Imp"
apply (simp add: Spc_def Imp_def)

  (***** by fixed point induction *****)
apply (rule cspF_fp_induct_left[of _ "Spc_to_Imp"])

  (***** check guarded and no hiding operators *****)

apply (simp_all)

  (***** by step laws (for transforming it to a hnf) *****)

apply (induct_tac p)
apply (tactic {* cspF_hsf_tac 1 *})+
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})+

  (***** instantiate a non-deterministic choice *****)

apply (rule cspF_Rep_int_choice_left)
apply (rule_tac  x="(Suc nat)" in exI)
apply (simp add: GTs_def)

apply (tactic {* cspF_unwind_tac 1 *})
apply (tactic {* cspF_light_step_tac 1 *})
done

end
