           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |                   December 2004 (modified)      |
            |                   November 2005 (modified)      |
            |                      April 2006 (modified)      |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_ccl = ep2_acl:

(*********************************************************
                    functions
 *********************************************************)

(* pair *)
consts
  state  :: "TerminalState * Trigger => TerminalState"
  trigger:: "TerminalState * Trigger => Trigger"

defs
  state_def  : "state p == fst p"
  trigger_def: "trigger p == snd p"

(* Session start *)
consts
  sessionStart:: "Trigger => D_SI_Init_SessionStart"

(* ConfigDataRequest *)
consts
  configDataResponse:: 
    "D_SI_Init_ConfigDataRequest * TerminalState => 
     D_SI_Init_ConfigDataResponse"

(* ConfigDataNotification *)
consts
  configDataAcknowledge:: D_SI_Init_ConfigDataAcknowledge 
  configData::
    "D_SI_Init_ConfigDataNotification * TerminalState =>
     TerminalState"

(* ConfigDataRemove *)
consts
  removeDataAcknowledge:: D_SI_Init_RemoveConfigDataAcknowledge
  removeData:: "D_SI_Init_RemoveConfigDataNotification * TerminalState =>
                TerminalState"

(* ActivateConfigDataNotification *)
consts
  activateDataAcknowledge:: D_SI_Init_ActivateConfigDataAcknowledge
  activateData:: "D_SI_Init_ActivateConfigDataNotification * TerminalState =>
                  TerminalState"

(* Message *)
consts
  AcqConnectionFailed    :: Message
  InitialisationFinished :: Message
  InitialisationFailed   :: Message

(*********************************************************
         concrete component description level
 *********************************************************)

datatype CCName = CTInit "TerminalState * Trigger"
                | CTConfigurationManagement "TerminalState * Trigger"
                | CAcquirerInit
                | CConfigurationManagement

consts 
  CCDef :: "(CCName, Event) procDef"

recdef CCDef "{}"
  "CCDef (X, CTInit p) 
    = C_SI_Init ! (SStart (sessionStart (trigger p)))
      -> (X (CTConfigurationManagement p))"

  "CCDef (X, CTConfigurationManagement p)
    = C_SI_Init ? x ->
       IF (x:range CDReq)
       THEN C_SI_Init ! 
               (CDRes (configDataResponse 
               (((inv CDReq) x),(state p)))) 
            -> (X (CTConfigurationManagement p))

       ELSE IF (x:range CDN)
       THEN C_SI_Init ! (CDA configDataAcknowledge)
            -> (X (CTConfigurationManagement 
                ((configData (((inv CDN) x),(state p))), (trigger p))))

       ELSE IF (x:range RCDN)
       THEN C_SI_Init ! (RCDA removeDataAcknowledge)
            -> (X (CTConfigurationManagement 
                (removeData((inv RCDN) x,(state p)), (trigger p))))

       ELSE IF (x:range ACDN)
       THEN C_SI_Init ! (ACDA activateDataAcknowledge)
            -> (X (CTConfigurationManagement 
                (activateData((inv ACDN) x,(state p)), (trigger p))))

       ELSE IF (x:range SEnd)
       THEN C_TerminalDisplay ! InitialisationFinished -> SKIP
       ELSE STOP"

  "CCDef (X, CAcquirerInit)
    = C_SI_Init ? sessionStart: (range SStart)
      -> (X CConfigurationManagement)"

  "CCDef (X, CConfigurationManagement)
    =     C_SI_Init !? sessionEnd:(range SEnd) -> SKIP
      |~| C_SI_Init !? request:(range CDReq)
          -> C_SI_Init ? response:(range CDRes)
          -> (X CConfigurationManagement)
      |~| C_SI_Init !? notification:(range CDN)
          -> C_SI_Init ? acknowledge:(range CDA)
          -> (X CConfigurationManagement)
      |~| C_SI_Init !? notification:(range RCDN)
          -> C_SI_Init ? acknowledge:(range RCDA)
          -> (X CConfigurationManagement)
      |~| C_SI_Init !? notification:(range ACDN)
          -> C_SI_Init ? acknowledge:(range ACDA)
          -> (X CConfigurationManagement)"

syntax
  "_CCFun" :: "(CCName, Event) procFun" ("CCFun")
  "_CCFix" :: "CCName => Event proc"    ("CCFix")

translations
    "CCFun" == "curry CCDef"
  "CCFix n" == "(FIX! CCFun) n"

consts
  CC :: "(TerminalState * Trigger) => Event proc"

defs
  CC_def: "CC p == ((CCFix (CTInit p))-- (range C_TerminalDisplay)
                   |[range C_SI_Init]| CCFix CAcquirerInit)"

(*********************************************************
                gProc lemmas (routine work)
 *********************************************************)

lemma CC_gProcFun[simp]:
      "CCFun : gProcFun"
by (simp add: CSP_Fun_def, rule allI, induct_tac p, simp_all)+

(*********************************************************
        relating function between AbsName and ACName
 *********************************************************)

consts 
  AC_to_CC :: "ACName => Event proc"

primrec
  "AC_to_CC (TInit)                    =  
     !<PairTT> p .. (CCFix (CTInit p))-- range C_TerminalDisplay"

  "AC_to_CC (TConfigurationManagement) = 
     !<PairTT> p ..  (CCFix (CTConfigurationManagement p)) -- range C_TerminalDisplay"

  "AC_to_CC (AcquirerInit)             = CCFix CAcquirerInit"
  "AC_to_CC (ConfigurationManagement)  = CCFix CConfigurationManagement"

(*********************************************************
           a theorem for verifying !!p. AC <=F CC p
 *********************************************************)

declare inj_on_def [simp]

lemma ep2_ccl_terminal_step1:
    "ACDef (AC_to_CC, TInit) <=F AC_to_CC TInit"
apply (auto)
apply (tactic {* cspF_hsf_tac 1 *})+
apply (simp add: image_iff)
apply (tactic {* cspF_simp_tac 1 *})+
apply (rule cspF_decompo_subset)
apply (auto)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="(a,b)" in exI)
apply (simp)
done

(*** loop (terminal) ***)

lemma ep2_ccl_terminal_step2:
    "ACDef (AC_to_CC, TConfigurationManagement) <=F 
     AC_to_CC TConfigurationManagement"
apply (auto)
apply (tactic {* cspF_unwind_tac 1 *})
apply (tactic {* cspF_step_tac 1 *})+
apply (simp add: image_iff)
apply (subgoal_tac "range C_SI_Init Int range C_TerminalDisplay = {}")
apply (tactic {* cspF_simp_tac 1 *})
apply (auto)

apply (case_tac "x : range CDReq")
apply (subgoal_tac "(EX xa. x = CDReq xa)", erule exE)
apply (tactic {* cspF_step_tac 1 *})+
apply (subgoal_tac "(C_SI_Init (CDRes (configDataResponse (xa, state (a, b))))
                    ~: range C_TerminalDisplay)")
apply (tactic {* cspF_simp_tac 1 *})
apply (rule cspF_decompo_subset)
apply (auto)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="(a, b)" in exI)
apply (simp)

apply (case_tac "(EX xa. x = CDReq xa)")
apply (auto)
apply (tactic {* cspF_step_tac 1 *})+

apply (case_tac "(EX xa. x = CDN xa)")
apply (auto)
apply (tactic {* cspF_step_tac 1 *})+

apply (subgoal_tac "(C_SI_Init (CDA configDataAcknowledge) ~: range C_TerminalDisplay)")
apply (tactic {* cspF_simp_tac 1 *})
apply (auto)

apply (rule cspF_decompo_subset)
apply (auto)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="(configData (xa, state (a, b)), trigger (a, b))" in exI)
apply (simp)

apply (tactic {* cspF_simp_tac 1 *})
apply (case_tac "(EX xa. x = RCDN xa) ")
apply (auto)
apply (tactic {* cspF_step_tac 1 *})+
apply (subgoal_tac "(C_SI_Init (RCDA removeDataAcknowledge) ~: range C_TerminalDisplay)")
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})

apply (rule cspF_decompo_subset)
apply (auto)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="(removeData (xa, state (a, b)), trigger (a, b))" in exI)
apply (simp)

apply (tactic {* cspF_simp_tac 1 *})
apply (case_tac "(EX xa. x = ACDN xa)")
apply (auto)
apply (tactic {* cspF_step_tac 1 *})+
apply (subgoal_tac "(C_SI_Init (ACDA activateDataAcknowledge) ~: range C_TerminalDisplay)")
apply (auto)
apply (tactic {* cspF_simp_tac 1 *})+

apply (rule cspF_decompo_subset)
apply (auto)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="(activateData (xa, state (a, b)), trigger (a, b))" in exI)
apply (simp)

 (*** C_TerminalDisplay ***)

apply (tactic {* cspF_simp_tac 1 *})
apply (case_tac "(EX xa. x = SEnd xa)")
apply (auto)
apply (tactic {* cspF_step_tac 1 *})+
done

 (*** ACDef_AC_to_CC ***)

lemma ACDef_AC_to_CC: "ACDef (AC_to_CC, p) <=F AC_to_CC p"
apply (induct_tac p)
apply (simp add: ep2_ccl_terminal_step1 del: ACDef.simps AC_to_CC.simps)
apply (simp add: ep2_ccl_terminal_step2 del: ACDef.simps AC_to_CC.simps)
apply (tactic {* cspF_unwind_tac 1 *})+
done

(****************************
      !!p. AC p <=F CC p
 ****************************)

theorem ep2_acl_ccl: "!!p. AC <=F CC p"
apply (simp add: AC_def CC_def)
apply (rule cspF_decompo)
apply (simp)

(*** TInit ***)

 (* by fixed point induction *)

apply (rule cspF_fp_induct_left[of _ "AC_to_CC"])
apply (simp_all)
apply (rule cspF_Rep_int_choice_left)
apply (simp)
apply (rule_tac x="p" in exI)
apply (simp)

apply (simp add: ACDef_AC_to_CC)

(*** AcquirerInit ***)

 (* by fixed point induction *)

apply (rule cspF_fp_induct_left[of _ "AC_to_CC"])
apply (simp_all)
apply (simp add: ACDef_AC_to_CC)
done

(****************************
      !!p. Abs <=F CC p
 ****************************)

theorem ep2_abs_ccl: "!!p. Abs <=F CC p"
apply (rule cspF_trans)
apply (rule ep2_abs)
apply (rule ep2_acl_ccl)
done

declare inj_on_def [simp del]

end
