           (*-------------------------------------------------*
            |                 (a part of) ep2                 |
            |                  September 2004                 |
            |                   December 2004 (modified)      |
            |                   November 2005 (modified)      |
            |                      April 2006 (modified)      |
            |  Markus Roggenbach (Univ of Wales Swansea, UK)  |
            |  Yoshinao Isobe    (AIST, Japan)                |
            *-------------------------------------------------*)

theory ep2_nucleus = CSP_F:

(*** To automatically unfold syntactic sugar ***)
declare csp_prefix_ss_def [simp]

(*********************************************************
              data type passed on channels
 *********************************************************)

typedecl init_d typedecl request_d typedecl response_d typedecl exit_d

datatype Data = Init    init_d    | Exit     exit_d
              | Request request_d | Response response_d

(*********************************************************
                     event (channel)
 *********************************************************)

datatype Event = c Data

(*********************************************************
         abstract component description level
 *********************************************************)

datatype ACName = Acquirer | AcConfigManagement 
                | Terminal | TerminalConfigManagement

consts 
  ACDef :: "(ACName, Event) procDef"

recdef ACDef "{}"
  "ACDef (X, Acquirer)
         = c ? x:(range Init) -> (X AcConfigManagement)"

  "ACDef (X, AcConfigManagement)
         = c !? exit:(range Exit) -> SKIP
           |~| 
           c !? request:(range Request)
             -> c ? response:(range Response)
             -> (X AcConfigManagement)"

  "ACDef (X, Terminal)
         = c !? init:(range Init) -> (X TerminalConfigManagement)"

  "ACDef (X, TerminalConfigManagement)
         = c ? x -> IF (x:range Request) 
                    THEN c !? response:(range Response)
                          -> (X TerminalConfigManagement)
                    ELSE IF (x:range Exit)
                         THEN SKIP
                         ELSE STOP"

syntax
  "_ACFun" :: "(ACName, Event) procFun" ("ACFun")
  "_ACFix" :: "ACName => Event proc"    ("ACFix")

translations
    "ACFun" == "curry ACDef"
  "ACFix n" == "(FIX! ACFun) n"

consts 
  AC :: "Event proc"

defs
  AC_def: "AC == (ACFix Acquirer |[range c]| ACFix Terminal)"

(*---------------------------------------------------------------*
 |                          NOTE                                 |
 |                                                               |
 | c ! v -> P       : sends a value v to c, then behaves like P. |
 |                                                               |
 | c ? x:X -> P(x)  : receives a value v from c if c in X,       |
 |                    then behaves like P(v).                    |
 |                                                               |
 | c !? x:X -> P(x) : sends a value v selected from X to c,      |
 |                    then behaves like P(v).                    |
 |                                                               |
 *---------------------------------------------------------------*)

(*********************************************************
                  architectual level
 *********************************************************)

datatype AbsName = Abstract | Loop

consts 
  AbsDef :: "(AbsName, Event) procDef"

recdef AbsDef "{}"
  "AbsDef (X, Abstract)
          = c !? init:(range Init) -> (X Loop)"

  "AbsDef (X, Loop)
          = c !? exit:(range Exit) -> SKIP
            |~| 
            c !? request:(range Request)
              -> c !? response:(range Response)
              -> (X Loop)"

syntax
  "_AbsFun" :: "(AbsName, Event) procFun" ("AbsFun")
  "_AbsFix" :: "AbsName => Event proc"    ("AbsFix")

translations
    "AbsFun" == "curry AbsDef"
  "AbsFix n" == "(FIX! AbsFun) n"

consts 
  Abs :: "Event proc"

defs
  Abs_def: "Abs == (AbsFix Abstract)"

(*********************************************************
        relating function between ACName and AbsName
 *********************************************************)

consts 
  Abs_to_AC :: "AbsName => Event proc"

primrec
  "Abs_to_AC (Abstract)
          = (ACFix Acquirer |[range c]| ACFix Terminal)"

  "Abs_to_AC (Loop)
          = (ACFix AcConfigManagement |[range c]| ACFix TerminalConfigManagement)"

(*********************************************************
               gProc lemmas (routine work)
 *********************************************************)

lemma gProcFun_AC_Abd[simp]:
      "ACFun : gProcFun"
      "AbsFun : gProcFun"
by (simp add: CSP_Fun_def, rule allI, induct_tac p, simp_all)+

(*********************************************************
           a theorem for verifying Abs <=F AC
 *********************************************************)

theorem ep2: "Abs <=F AC"
apply (unfold Abs_def AC_def)
apply (rule cspF_fp_induct_left[of _ "Abs_to_AC"])
apply (simp_all)
apply (induct_tac p)
apply (simp_all)

(* 1st subgoal *)
apply (tactic {* cspF_hsf_tac 1 *})+
apply (rule cspF_decompo, auto)
apply (tactic {* cspF_hsf_tac 1 *})
apply (rule cspF_decompo, auto)
apply (tactic {* cspF_simp_tac 1 *})

(* 2nd subgoal *)
apply (tactic {* cspF_hsf_tac 1 *} | rule cspF_decompo | rule cspF_decompo_subset | 
       auto simp add: image_iff inj_on_def)+
done

end
